# -*- coding: utf-8 -*-

# Copyright 2018 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt caching tests
~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""

import json
from os.path import isdir, isfile, join

import pytest

from .caching import (get_prebuilt_targets, save_target_in_cache,
                      save_test_in_cache)
from .buildcontext import BuildContext
from .graph import populate_targets_graph
from .utils import rmtree


slow = pytest.mark.skipif(not pytest.config.getoption('--with-slow'),
                          reason='need --with-slow option to run')


@slow
@pytest.mark.usefixtures('in_caching_project')
def test_prebuilt_targets_case1(basic_conf):
    """Test pre-built case #1 - all base deps should be marked as pre-built.

    See issue: https://github.com/resonai/ybt/issues/61
    """
    basic_conf.targets = [':builder']
    build_context = BuildContext(basic_conf)
    populate_targets_graph(build_context, basic_conf)
    pre_built = get_prebuilt_targets(build_context)
    assert set((':build-tools', ':tools', ':unzip', ':ubuntu')) == pre_built
    assert (set((':builder', ':builder-base', ':build-tools',
                 ':tools', ':unzip', ':ubuntu')) ==
            set(build_context.target_graph.nodes))


@slow
@pytest.mark.usefixtures('in_caching_project')
def test_prebuilt_targets_case2(basic_conf):
    """Test pre-built case #2 - nothing should be marked pre-built.

    See issue: https://github.com/resonai/ybt/issues/61
    """
    basic_conf.targets = [':an-image']
    build_context = BuildContext(basic_conf)
    populate_targets_graph(build_context, basic_conf)
    pre_built = get_prebuilt_targets(build_context)
    assert set() == pre_built
    assert (set((':an-image', ':unzip', ':ubuntu')) ==
            set(build_context.target_graph.nodes))


@slow
@pytest.mark.usefixtures('in_caching_project')
def test_prebuilt_targets_case1(basic_conf):
    """Test pre-built case #3 - unzip & ubuntu should NOT mark as prebuilt.

    See issue: https://github.com/resonai/ybt/issues/61
    """
    basic_conf.targets = [':all-images']
    build_context = BuildContext(basic_conf)
    populate_targets_graph(build_context, basic_conf)
    pre_built = get_prebuilt_targets(build_context)
    assert set((':build-tools', ':tools')) == pre_built
    assert (set((':all-images', ':an-image', ':builder', ':builder-base',
                 ':build-tools', ':tools', ':unzip', ':ubuntu')) ==
            set(build_context.target_graph.nodes))


@slow
@pytest.mark.usefixtures('in_caching_project')
def test_prebuilt_targets_build_base_image(basic_conf):
    """Test pre-built targets when building base images.

    When `--build-base-images` is specified, all targets should be built,
    regardless of base-image status.
    """
    basic_conf.build_base_images = True
    basic_conf.targets = [':builder']
    build_context = BuildContext(basic_conf)
    populate_targets_graph(build_context, basic_conf)
    pre_built = get_prebuilt_targets(build_context)
    assert set() == pre_built
    assert (set((':builder', ':builder-base', ':build-tools',
                 ':tools', ':unzip', ':ubuntu')) ==
            set(build_context.target_graph.nodes))


def read_file(fpath, parse_json=False):
    with open(fpath, 'r') as f:
        content = f.read()
    if parse_json:
        return json.loads(content)
    return content


_EXP_UNZIP_JSON = """{
    "buildenv": [],
    "builder_name": "AptPackage",
    "deps": [],
    "flavor": null,
    "name": ":unzip",
    "props": {
        "attempts": 1,
        "build_params": {},
        "license": [],
        "package": "unzip",
        "packaging_params": {},
        "policies": [],
        "repo_key": null,
        "repo_keyserver": "hkp://keyserver.ubuntu.com:80",
        "repository": null,
        "runtime_params": {},
        "version": null
    },
    "tags": [
        "apt-installable"
    ]
}"""


@pytest.mark.usefixtures('in_caching_project')
def test_save_target_to_cache(basic_conf):
    cache_dir = join(basic_conf.project_root, 'yabtwork', '.cache')
    rmtree(cache_dir)
    basic_conf.targets = [':all-images']
    build_context = BuildContext(basic_conf)
    populate_targets_graph(build_context, basic_conf)
    target_name = ':unzip'
    unzip_target = build_context.targets[target_name]
    unzip_target.summary['build_time'] = 5.432
    save_target_in_cache(unzip_target, build_context)
    target_cache_dir = join(cache_dir, 'targets',
                            unzip_target.hash(build_context))
    assert isdir(target_cache_dir)
    target_json_path = join(target_cache_dir, 'target.json')
    artifacts_json_path = join(target_cache_dir, 'artifacts.json')
    summary_json_path = join(target_cache_dir, 'summary.json')
    assert isfile(target_json_path)
    assert isfile(artifacts_json_path)
    assert isfile(summary_json_path)
    assert _EXP_UNZIP_JSON == read_file(target_json_path)
    assert '{}' == read_file(artifacts_json_path)
    summary = read_file(summary_json_path, True)
    assert set(('accessed', 'artifacts_hash', 'build_time', 'created',
                'name')) == set(summary.keys())
    assert summary['build_time'] == 5.432
    assert summary['name'] == target_name


@pytest.mark.usefixtures('in_caching_project')
def test_save_test_to_cache(basic_conf):
    cache_dir = join(basic_conf.project_root, 'yabtwork', '.cache')
    rmtree(cache_dir)
    basic_conf.targets = [':all-images']
    build_context = BuildContext(basic_conf)
    populate_targets_graph(build_context, basic_conf)
    target_name = ':unzip'
    unzip_target = build_context.targets[target_name]
    unzip_target.summary['build_time'] = 5.432
    # If there's no test to cache, the caching should always succeed.
    assert save_test_in_cache(unzip_target, build_context)
    unzip_target.tested.update({'some_test': 5.434})
    # Cannot cache a test without first caching the target.
    assert not save_test_in_cache(unzip_target, build_context)
    # If the build was cached first, the test can also be cached.
    save_target_in_cache(unzip_target, build_context)
    assert save_test_in_cache(unzip_target, build_context)
    target_cache_dir = join(cache_dir, 'targets',
                            unzip_target.hash(build_context))
    assert isdir(target_cache_dir)
    target_json_path = join(target_cache_dir, 'target.json')
    artifacts_json_path = join(target_cache_dir, 'artifacts.json')
    summary_json_path = join(target_cache_dir, 'summary.json')
    tested_json_path = join(target_cache_dir, 'tested.json')
    assert isfile(target_json_path)
    assert isfile(artifacts_json_path)
    assert isfile(summary_json_path)
    assert isfile(tested_json_path)



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-argument

"""
yabt Apt Builders
~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


import requests

from ..extend import (
    PropType as PT, register_build_func, register_builder_sig,
    register_manipulate_target_hook)
from ..utils import yprint


LAUNCHPAD_URL = ('https://launchpad.net/api/1.0/'
                 '~{ppa_owner}/+archive/{ppa_name}')
LAUNCHPAD_SOURCE_LINE = ('deb http://ppa.launchpad.net/{ppa_owner}/{ppa_name}/'
                         '{distro_id} {distro_codename} main')
VALID_SOURCE_TYPES = frozenset(('deb',))  # 'deb-src'


register_builder_sig(
    'AptPackage',
    [('package', PT.str),
     ('version', PT.str, None),
     ('repository', PT.str, None),
     ('repo_key', PT.str, None),
     ('repo_keyserver', PT.str, 'hkp://keyserver.ubuntu.com:80'),
     ])


@register_build_func('AptPackage')
def apt_package_builder(build_context, target):
    yprint(build_context.conf, 'Fetch and cache Apt package', target)


@register_manipulate_target_hook('AptPackage')
def apt_package_manipulate_target(build_context, target):
    target.tags.add('apt-installable')
    if target.props.repository:
        target.tags.add('apt-repository')
        # translate to AptRepository props
        target.props.source = target.props.pop('repository')
        target.props.key = target.props.pop('repo_key')
        target.props.keyserver = target.props.pop('repo_keyserver')


register_builder_sig(
    'AptGroup',
    [('packages', PT.list),
     ('repository', PT.str, None),
     ('repo_key', PT.str, None),
     ('repo_keyserver', PT.str, 'hkp://keyserver.ubuntu.com:80'),
     ])


@register_build_func('AptGroup')
def apt_group_builder(build_context, target):
    yprint(build_context.conf, 'Fetch and cache Apt packages', target)


@register_manipulate_target_hook('AptGroup')
def apt_group_manipulate_target(build_context, target):
    target.tags.add('apt-installable')
    if target.props.repository:
        target.tags.add('apt-repository')
        # translate to AptRepository props
        target.props.source = target.props.pop('repository')
        target.props.key = target.props.pop('repo_key')
        target.props.keyserver = target.props.pop('repo_keyserver')


register_builder_sig(
    'AptRepository',
    [('source', PT.str),
     ('key', PT.str, None),
     ('keyserver', PT.str, 'hkp://keyserver.ubuntu.com:80'),
     ])


@register_build_func('AptRepository')
def apt_repository_builder(build_context, target):
    pass


@register_manipulate_target_hook('AptRepository')
def apt_repository_manipulate_target(build_context, target):
    target.tags.add('apt-repository')



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-argument

"""
yabt Python Builders
~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""

from itertools import chain
from os.path import join, relpath

from ostrich.utils.collections import listify

from ..artifact import ArtifactType as AT
from .dockerapp import build_app_docker_and_bin, register_app_builder_sig
from ..extend import (
    PropType as PT, register_build_func, register_builder_sig,
    register_manipulate_target_hook, register_test_func)
from ..logging import make_logger
from ..utils import rmtree, yprint


logger = make_logger(__name__)


def path_to_pymodule(fpath):
    if fpath.endswith('.py'):
        fpath = fpath[:-3]
    return '.'.join(fpath.split('/'))


register_builder_sig(
    'PythonPackage',
    [('package', PT.str),
     ('version', PT.str, None),
     ('pip', PT.str, 'pip2'),
     ])


@register_build_func('PythonPackage')
def python_package_builder(build_context, target):
    yprint(build_context.conf, 'Fetch and cache PyPI package', target)


@register_manipulate_target_hook('PythonPackage')
def python_package_manipulate_target(build_context, target):
    target.tags.add('pip-installable')


register_builder_sig(
    'Python',
    [('sources', PT.FileList, None),
     ('data', PT.FileList, None),
     ])


@register_build_func('Python')
def python_app_builder(build_context, target):
    yprint(build_context.conf, 'Build Python', target)
    for src in chain(target.props.sources, target.props.data):
        target.artifacts.add(AT.app, src)


register_app_builder_sig('PythonApp', [('main', PT.File)])


@register_manipulate_target_hook('PythonApp')
def python_app_manipulate_target(build_context, target):
    logger.debug('Injecting {} to deps of {}',
                 target.props.base_image, target.name)
    target.deps.append(target.props.base_image)


@register_build_func('PythonApp')
def python_app_builder(build_context, target):
    yprint(build_context.conf, 'Build PythonApp', target)
    build_app_docker_and_bin(
        build_context, target, entrypoint=[target.props.main])


register_builder_sig(
    'PythonTest',
    [('module', PT.File),
     ('test_cmd', PT.StrList, None),
     ('test_flags', PT.StrList, None),  # flags to append to test command
     ('test_env', None),  # env vars to inject in test process
     ('in_testenv', PT.Target, None),
     ])


@register_manipulate_target_hook('PythonTest')
def pythontest_manipulate_target(build_context, target):
    target.tags.add('testable')
    target.buildenv = target.props.in_testenv
    # doing test-command manipulation (and storing it in target props) during
    # target extraction (as opposed to during build) so it is considered
    # during target hashing (for cache)
    prepare_test_cmd(build_context, target)


@register_build_func('PythonTest')
def pythontest_builder(build_context, target):
    yprint(build_context.conf, 'Build PythonTest', target)


@register_test_func('PythonTest')
def pythontest_tester(build_context, target):
    """Run a Python test"""
    yprint(build_context.conf, 'Run PythonTest', target)
    workspace_dir = build_context.get_workspace('PythonTest', target.name)

    # Link generated Python protos & required binaries
    gen_dir = join(workspace_dir, 'gen')
    rmtree(gen_dir)
    for dep in build_context.generate_all_deps(target):
        dep.artifacts.link_types(
            workspace_dir, [AT.gen_py, AT.binary], build_context.conf)

    # Run the test module
    test_env = target.props.test_env or {}
    pypath = test_env.get('PYTHONPATH', '.').split(':')
    pypath.append(relpath(gen_dir, build_context.conf.project_root))
    test_env['PYTHONPATH'] = ':'.join(pypath)
    test_env['BIN_DIR'] = build_context.conf.host_to_buildenv_path(
        join(workspace_dir, 'bin'))
    build_context.run_in_buildenv(
        target.props.in_testenv, target.props.test_cmd, test_env)


def prepare_test_cmd(build_context, target):
    pytest_params = build_context.conf.get('pytest_params', {})
    test_cmd = target.props.test_cmd
    if not test_cmd:
        test_cmd = list(listify(pytest_params.get('default_test_cmd',
                                                  ['python', '-m'])))
    test_cmd.append(path_to_pymodule(target.props.module))
    test_cmd.extend(target.props.test_flags)
    test_cmd.extend(listify(pytest_params.get('extra_exec_flags')))
    target.props.test_cmd = test_cmd



# -*- coding: utf-8 -*-

# Copyright 2019 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
dot generation tests
~~~~~~~~~~~~~~~~~~~~~~~

:author: Dana Shamir
"""


import io
import pytest
import re

from yabt.buildcontext import BuildContext
from yabt.dot import TARGETS_COLORS, write_dot, get_not_buildenv_targets
from yabt.graph import populate_targets_graph


slow = pytest.mark.skipif(not pytest.config.getoption('--with-slow'),
                          reason='need --with-slow option to run')


@pytest.mark.usefixtures('in_dag_project')
def test_graph_dot_generation(basic_conf):
    build_context = BuildContext(basic_conf)
    populate_targets_graph(build_context, basic_conf)
    expected_targets = {':flask', ':gunicorn', 'common:logging', 'common:base',
                        'fe:fe', 'yapi/server:users', 'yapi/server:yapi',
                        'yapi/server:yapi-gunicorn'}
    expected_dot_nodes = {'  "{}" [color="black",];'.format(target)
                          for target in expected_targets}
    expected_dot_edges = set([
        '  "common:base" -> "common:logging";',
        '  "fe:fe" -> "yapi/server:users";', '  "fe:fe" -> "common:base";',
        '  "fe:fe" -> ":flask";', '  "yapi/server:yapi" -> "common:base";',
        '  "yapi/server:yapi" -> ":flask";',
        '  "yapi/server:yapi-gunicorn" -> "yapi/server:yapi";',
        '  "yapi/server:yapi-gunicorn" -> "common:base";',
        '  "yapi/server:yapi-gunicorn" -> ":gunicorn";'])
    with io.StringIO() as dot_io:
        write_dot(build_context, basic_conf, dot_io)
        dot_lines = dot_io.getvalue().strip('\n').split('\n')
        assert 'strict digraph  {' == dot_lines[0]
        assert '}' == dot_lines[-1]
        assert expected_dot_nodes == set(dot_lines[1:9])
        assert expected_dot_edges == set(dot_lines[9:18])


@pytest.mark.usefixtures('in_simpleflat_project')
def test_graph_dot_generation_colors(basic_conf):
    build_context = BuildContext(basic_conf)
    populate_targets_graph(build_context, basic_conf)
    expected_targets = {':flask-0.10.1': TARGETS_COLORS['PythonPackage'],
                        ':flask-hello-app': TARGETS_COLORS['Python']}
    expected_dot_nodes = ['  "{}" [color="{}",];'.format(target, color)
                          for target, color in expected_targets.items()]
    with io.StringIO() as dot_io:
        write_dot(build_context, basic_conf, dot_io)
        dot_lines = dot_io.getvalue().strip('\n').split('\n')
        for dot_node in expected_dot_nodes:
            assert dot_node in dot_lines


@pytest.mark.usefixtures('in_cpp_project')
def test_no_buildenv_deps_in_dot(basic_conf):
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['hello:hello-app']
    populate_targets_graph(build_context, basic_conf)
    buildenv_targets = {':builder', ':ubuntu-gpg', ':clang', ':gnupg'}
    expected_targets = {'hello:hello-app', 'hello:hello', ':ubuntu'}
    with io.StringIO() as dot_io:
        write_dot(build_context, basic_conf, dot_io)
        all_targets = set(dot_io.getvalue().split('"'))
        assert not buildenv_targets.intersection(all_targets)
        assert expected_targets.intersection(all_targets) == expected_targets


@slow
@pytest.mark.usefixtures('in_caching_project')
def test_cached_targets_different_color(basic_conf):
    basic_conf.targets = [':builder']
    build_context = BuildContext(basic_conf)
    populate_targets_graph(build_context, basic_conf)
    cached_targets = {':build-tools', ':tools', ':unzip', ':ubuntu'}
    other_targets = {':builder', ':builder-base'}
    expected_dot_nodes = set([
        '  "{}" \[color=".*",fillcolor="grey",style=filled\];'.format(target)
        for target in cached_targets] + [
        '  "{}" \[color=".*",\];'.format(target) for target in other_targets])
    with io.StringIO() as dot_io:
        write_dot(build_context, basic_conf, dot_io)
        dot_lines = dot_io.getvalue().strip('\n').split('\n')
        assert set(filter(lambda expected:
                          any(filter(lambda line: re.match(expected, line),
                                     dot_lines[1:7])),
                          expected_dot_nodes)) == expected_dot_nodes


@pytest.mark.usefixtures('in_cpp_project')
def test_buildenv_and_target_dep(basic_conf):
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['hello_lib:hello-gnupg']
    populate_targets_graph(build_context, basic_conf)
    assert ':gnupg' in get_not_buildenv_targets(build_context)



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-argument

"""
yabt C++ Builder
~~~~~~~~~~~~~~~~

:author: Itamar Ostricher

TODO: support injecting compile/link flags for 3rd party libs
TODO: CppSharedLib builder
"""


from os.path import basename, dirname, join, relpath, splitext

from ostrich.utils.collections import listify

from ..artifact import ArtifactType as AT
from .dockerapp import build_app_docker_and_bin, register_app_builder_sig
from ..extend import (
    PropType as PT, register_build_func, register_builder_sig,
    register_manipulate_target_hook, register_test_func)
from ..logging import make_logger
from ..target_utils import split
from ..utils import link_files, rmtree, yprint


logger = make_logger(__name__)


class CompilerConfig:
    """Helper class for managing compiler / linker options and flags.

    Design note:
    Notice the `extra_{compile,link}_flags` for target X are collected from
    all the dependencies of target X, but are actually applied to a command
    that is executed within the **build-env image** of target X, which can
    have a completely unrelated subgraph!
    It is the responsibility of the "user" to use these parameters correctly,
    and make sure that the build-env-image contains the **build-time**
    dependencies that are needed to make this parameters work *there*, while
    the target itself has the correct **runtime** dependencies.
    See an example of how this plays out in the YaBT tests - specifically,
    check out the `cpp/hello_boost` test and examine how the build-env image
    contains boost dev-lib (for build-time, including all boost sub-libs),
    while the final target depends on boost-runtime, which installs only 3
    boost shared objects that are needed for this specific application.

    TODO: add this note to proper project docs (also - create proper docs...)
    """

    def __init__(self, build_context, target, extra_params=None):
        self.compiler = self.get(
            'compiler', build_context.conf, target, 'g++')
        self.linker = self.get(
            'linker', build_context.conf, target, 'g++')
        self.compile_flags = list(self.get(
            'compile_flags', build_context.conf, target, []))
        self.link_flags = list(self.get(
            'link_flags', build_context.conf, target, []))
        self.include_path = list(self.get(
            'include_path', build_context.conf, target, []))

        def generate_extra_params():
            if extra_params:
                yield extra_params
            yield target.props.build_params
            # using topological order here because linker `-l<lib>` flags
            # are sensitive to the order that they appear in!
            # so if this target depends on, for example, both libsoft and
            # libfftw, and also libsoft requires symbols that are defined in
            # libfftw, then `-lfftw` must appear *after* `-lsoft`.
            yield from (
                dep.props.build_params for dep in
                build_context.walk_target_deps_topological_order(target))

        for build_params in reversed(list(generate_extra_params())):
            self.compile_flags.extend(
                listify(build_params.get('extra_compile_flags')))
            self.link_flags.extend(
                listify(build_params.get('extra_link_flags')))

    def as_dict(self):
        return {key: getattr(self, key)
                for key in ('compiler', 'linker', 'compile_flags',
                            'link_flags', 'include_path')}

    def get(self, param, config, target, fallback):
        """Return the value of `param`, according to priority / expansion.

        First priority - the target itself.
        Second priority - the project config.
        Third priority - a global default ("fallback").

        In list-params, a '$*' term processed as "expansion term", meaning
        it is replaced with all terms from the config-level.
        """
        target_val = target.props.get(param)
        config_val = config.get(param, fallback)
        if not target_val:
            return config_val
        if isinstance(target_val, list):
            val = []
            for el in target_val:
                if el == '$*':
                    val.extend(listify(config_val))
                else:
                    val.append(el)
            return val
        return target_val


# Common C++ builder signature terms
CPP_SIG = [
     ('sources', PT.FileList),
     ('in_buildenv', PT.Target),
     ('headers', PT.FileList, None),
     ('protos', PT.TargetList, None),
     ('cmd_env', None),
     ('compiler', PT.str, None),
     ('linker', PT.str, None),
     ('compile_flags', PT.StrList, None),
     ('link_flags', PT.StrList, None),
     ('include_path', PT.StrList, None),
     # an internal "prop" used internally bt YBT to add compiler config data
     # to the target props, so it is considered by target hashing (for cache)
     ('_internal_dict_', PT.dict, None),
]


register_app_builder_sig(
    'CppApp', [('executable', PT.File, None), ('main', PT.Target, None)])


@register_manipulate_target_hook('CppApp')
def cpp_app_manipulate_target(build_context, target):
    logger.debug('Injecting {} to deps of {}',
                 target.props.base_image, target.name)
    target.deps.append(target.props.base_image)
    if target.props.main and target.props.main not in target.deps:
        logger.debug('Injecting {} to deps of {}',
                     target.props.main, target.name)
        target.deps.append(target.props.main)


@register_build_func('CppApp')
def cpp_app_builder(build_context, target):
    """Pack a C++ binary as a Docker image with its runtime dependencies.

    TODO(itamar): Dynamically analyze the binary and copy shared objects
    from its buildenv image to the runtime image, unless they're installed.
    """
    yprint(build_context.conf, 'Build CppApp', target)
    if target.props.executable and target.props.main:
        raise KeyError(
            '`main` and `executable` arguments are mutually exclusive')
    if target.props.executable:
        if target.props.executable not in target.artifacts.get(AT.app):
            target.artifacts.add(AT.app, target.props.executable)
        entrypoint = [target.props.executable]
    elif target.props.main:
        prog = build_context.targets[target.props.main]
        binary = list(prog.artifacts.get(AT.binary).keys())[0]
        entrypoint = ['/usr/src/bin/' + binary]
    else:
        raise KeyError('Must specify either `main` or `executable` argument')
    build_app_docker_and_bin(
        build_context, target, entrypoint=entrypoint)


register_builder_sig('CppProg', CPP_SIG)
register_builder_sig(
    'CppGTest',
    CPP_SIG + [
        ('test_flags', PT.StrList, None),  # flags to append to test command
        ('test_env', None),  # env vars to inject in test process
        ('runtime', PT.str, None),
        # TODO: support different testenv image for test execution
        # ('in_testenv', PT.Target, None),
    ]
)


def make_pre_build_hook(extra_compiler_config_params):
    """Return a pre-build hook function for C++ builders.

    When called, during graph build, it computes and stores the compiler-config
    object on the target, as well as adding it to the internal_dict prop for
    hashing purposes.
    """

    def pre_build_hook(build_context, target):
        target.compiler_config = CompilerConfig(
            build_context, target, extra_compiler_config_params)
        target.props._internal_dict_['compiler_config'] = (
            target.compiler_config.as_dict())

    return pre_build_hook


@register_manipulate_target_hook('CppProg')
def cpp_prog_manipulate_target(build_context, target):
    target.buildenv = target.props.in_buildenv
    target.pre_build_hook = make_pre_build_hook({})


@register_manipulate_target_hook('CppGTest')
def cpp_gtest_manipulate_target(build_context, target):
    target.tags.add('testable')
    target.buildenv = target.props.in_buildenv
    # target.buildenvs.append(target.props.in_testenv)
    # manipulate the test_flags prop during target extraction (as opposed to
    # during build func), so it is considered during target hashing (for cache)
    target.props.test_flags.extend(listify(
        build_context.conf.get('gtest_params', {}).get('extra_exec_flags')))
    target.pre_build_hook = make_pre_build_hook(
        build_context.conf.get('gtest_params', {}))


def is_cc_file(filename: str) -> bool:
    return splitext(filename)[-1].lower() in ('.cc', '.cpp', '.cxx', '.c++')


def is_h_file(filename: str) -> bool:
    return splitext(filename)[-1].lower() in ('.h', '.hpp', '.hh', '.hxx')


def compile_cc(build_context, compiler_config, buildenv, sources,
               workspace_dir, buildenv_workspace, cmd_env):
    """Compile list of C++ source files in a buildenv image
       and return list of generated object file.
    """
    objects = []
    for src in sources:
        obj_rel_path = '{}.o'.format(splitext(src)[0])
        obj_file = join(buildenv_workspace, obj_rel_path)
        include_paths = [buildenv_workspace] + compiler_config.include_path
        compile_cmd = (
            [compiler_config.compiler, '-o', obj_file, '-c'] +
            compiler_config.compile_flags +
            ['-I{}'.format(path) for path in include_paths] +
            [join(buildenv_workspace, src)])
        # TODO: capture and transform error messages from compiler so file
        # paths match host paths for smooth(er) editor / IDE integration
        build_context.run_in_buildenv(buildenv, compile_cmd, cmd_env)
        objects.append(
            join(relpath(workspace_dir, build_context.conf.project_root),
                 obj_rel_path))
    return objects


def link_cpp_artifacts(build_context, target, workspace_dir,
                       include_objects: bool):
    """Link required artifacts from dependencies under target workspace dir.
       Return list of object files of dependencies (if `include_objects`).

    Includes:
    - Generated code from proto dependencies
    - Header files from all dependencies
    - Generated header files from all dependencies
    - If `include_objects` is True, also object files from all dependencies
      (these will be returned without linking)
    """
    # include the source & header files of the current target
    # add objects of all dependencies (direct & transitive), if needed
    source_files = target.props.sources + target.props.headers
    generated_srcs = {}
    objects = []

    # add headers of dependencies
    for dep in build_context.generate_all_deps(target):
        source_files.extend(dep.props.get('headers', []))

    link_files(source_files, workspace_dir, None, build_context.conf)

    # add generated headers and collect objects of dependencies
    for dep in build_context.generate_all_deps(target):
        dep.artifacts.link_types(workspace_dir, [AT.gen_h], build_context.conf)
        if include_objects:
            objects.extend(dep.artifacts.get(AT.object).values())

    # add generated code from proto dependencies
    for proto_dep_name in target.props.protos:
        proto_dep = build_context.targets[proto_dep_name]
        proto_dep.artifacts.link_types(workspace_dir, [AT.gen_cc],
                                       build_context.conf)

    return objects


def get_source_files(target, build_context) -> list:
    """Return list of source files for `target`."""
    all_sources = list(target.props.sources)
    for proto_dep_name in target.props.protos:
        proto_dep = build_context.targets[proto_dep_name]
        all_sources.extend(proto_dep.artifacts.get(AT.gen_cc).keys())
    return all_sources


def build_cpp(build_context, target, compiler_config, workspace_dir):
    """Compile and link a C++ binary for `target`."""
    rmtree(workspace_dir)
    binary = join(*split(target.name))
    objects = link_cpp_artifacts(build_context, target, workspace_dir, True)
    buildenv_workspace = build_context.conf.host_to_buildenv_path(
        workspace_dir)
    objects.extend(compile_cc(
        build_context, compiler_config, target.props.in_buildenv,
        get_source_files(target, build_context), workspace_dir,
        buildenv_workspace, target.props.cmd_env))
    bin_file = join(buildenv_workspace, binary)
    link_cmd = (
        [compiler_config.linker, '-o', bin_file] +
        objects + compiler_config.link_flags)
    build_context.run_in_buildenv(
        target.props.in_buildenv, link_cmd, target.props.cmd_env)
    target.artifacts.add(AT.binary, relpath(join(workspace_dir, binary),
                         build_context.conf.project_root), binary)


@register_build_func('CppProg')
def cpp_prog_builder(build_context, target):
    """Build a C++ binary executable"""
    yprint(build_context.conf, 'Build CppProg', target)
    workspace_dir = build_context.get_workspace('CppProg', target.name)
    build_cpp(build_context, target, target.compiler_config, workspace_dir)

    # TODO: remove?
    # # Copy binary artifacts to external destination
    # if target.props.copy_bin_to:
    #     link_artifacts([join(workspace_dir, binary)],
    #                    target.props.copy_bin_to,
    #                    workspace_dir, build_context.conf)


@register_build_func('CppGTest')
def cpp_gtest_builder(build_context, target):
    """Build a C++ test executable"""
    yprint(build_context.conf, 'Build CppGTest', target)
    workspace_dir = build_context.get_workspace('CppGTest', target.name)
    build_cpp(build_context, target, target.compiler_config, workspace_dir)


@register_test_func('CppGTest')
def cpp_gtest_tester(build_context, target):
    """Run a C++ test executable"""
    yprint(build_context.conf, 'Run CppGTest', target)
    workspace_dir = build_context.get_workspace('CppGTest', target.name)
    buildenv_workspace = build_context.conf.host_to_buildenv_path(
        workspace_dir)
    test_cmd = [join(buildenv_workspace, *split(target.name))]
    # take gtest exec flags from the target & from project config
    test_cmd.extend(target.props.test_flags)
    build_context.run_in_buildenv(
        # TODO: target.props.in_testenv,
        target.props.in_buildenv, test_cmd, target.props.test_env,
        runtime=target.props.runtime
    )


register_builder_sig('CppLib', CPP_SIG)


@register_manipulate_target_hook('CppLib')
def cpp_lib_manipulate_target(build_context, target):
    target.buildenv = target.props.in_buildenv
    # proto targets are also direct dependencies
    target.deps.extend(target.props.protos)
    target.pre_build_hook = make_pre_build_hook({})


@register_build_func('CppLib')
def cpp_lib_builder(build_context, target):
    """Build C++ object files"""
    yprint(build_context.conf, 'Build CppLib', target)
    workspace_dir = build_context.get_workspace('CppLib', target.name)
    workspace_src_dir = join(workspace_dir, 'src')
    rmtree(workspace_src_dir)
    link_cpp_artifacts(build_context, target, workspace_src_dir, False)
    buildenv_workspace = build_context.conf.host_to_buildenv_path(
        workspace_src_dir)
    objects = compile_cc(
        build_context, target.compiler_config, target.props.in_buildenv,
        get_source_files(target, build_context), workspace_src_dir,
        buildenv_workspace, target.props.cmd_env)
    for obj_file in objects:
        target.artifacts.add(AT.object, obj_file)



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt Logging
~~~~~~~~~~~~~~~~~~~~~~~

Implement logger adapter and brace formatter to allow using new-style string
formatting in logging messages, following the logging cookbook:
https://docs.python.org/3.5/howto/logging-cookbook.html#use-of-alternative-formatting-styles

:author: Itamar Ostricher
"""


import logging
import sys


class Message:  # pylint: disable=too-few-public-methods
    """Simple message class for logger adapter."""

    def __init__(self, fmt, args):
        self.fmt = fmt
        self.args = args

    def __str__(self):
        return self.fmt.format(*self.args)


class StyleAdapter(logging.LoggerAdapter):
    """A Style logger adapter to allow new-style string formatting in logging.

    Taken from Python logging cookbook:
    https://docs.python.org/3.5/howto/logging-cookbook.html#use-of-alternative-formatting-styles
    """

    def __init__(self, logger, extra=None):
        super().__init__(logger, extra or {})

    def log(self, level, msg, *args, **kwargs):
        if self.isEnabledFor(level):
            msg, kwargs = self.process(msg, kwargs)
            self.logger._log(level, Message(msg, args), (), **kwargs)  # noqa pylint: disable=protected-access


def add_stream_handler(logger, stream):
    """Add a brace-handler stream-handler using `stream` to `logger`."""
    handler = logging.StreamHandler(stream=stream)
    # Using Brace Formatter (see
    # https://docs.python.org/3.5/howto/logging-cookbook.html#use-of-alternative-formatting-styles)
    formatter = logging.Formatter(
        '{asctime} {name:24s} {levelname:8s} {message}', style='{')
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def configure_logging(conf):
    """Initialize and configure logging."""
    root_logger = logging.getLogger()
    root_logger.setLevel(getattr(logging, conf.loglevel.upper()))
    if conf.logtostderr:
        add_stream_handler(root_logger, sys.stderr)
    if conf.logtostdout:
        add_stream_handler(root_logger, sys.stdout)


def make_logger(name: str) -> logging.Logger:
    """Return a sub-logger with name `name`.

    Recommended use in top of using module:
    ```
    logger = logging.make_logger(__name__)
    ```
    """
    return StyleAdapter(logging.getLogger(name))



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt Target utils tests
~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""

from os.path import join

import pytest

from .buildcontext import BuildContext
from .graph import populate_targets_graph
from .target_utils import hashify_files, hashify_targets, norm_name


def test_norm_name_abs_ref():
    """Test norm_name abs ref to targets in nested modules."""
    assert 'cat/foo:bar' == norm_name('cat', '//cat/foo:bar')
    assert 'cat:bar' == norm_name('cat', '//cat:bar')
    assert 'dog:bark' == norm_name('cat', '//dog:bark')
    assert 'stuff:*' == norm_name('cat', '//stuff:*')
    assert 'stuff/**:*' == norm_name('cat', '//stuff/**:*')


def test_norm_name_abs_ref_top_level():
    """Test norm_name abs ref to targets in top level module."""
    assert ':spam' == norm_name('cat', '//:spam')
    assert ':*' == norm_name('cat', '//:*')
    assert '**:*' == norm_name('cat', '//**:*')


def test_norm_name_rel_ref_sub():
    """Test norm_name rel ref to sub-module."""
    assert 'cat/foo:bar' == norm_name('cat', 'foo:bar')
    assert 'cat/foo:bar' == norm_name('cat', './foo:bar')
    assert 'cat/foo:*' == norm_name('cat', 'foo:*')
    assert 'cat/foo/**:*' == norm_name('cat', 'foo/**:*')


def test_norm_name_rel_ref_local():
    """Test norm_name rel ref to same module."""
    assert 'cat:bar' == norm_name('cat', ':bar')
    assert 'cat:bar' == norm_name('cat', '.:bar')
    assert 'cat:bar' == norm_name('cat', './:bar')
    assert 'cat:*' == norm_name('cat', ':*')
    assert 'cat/**:*' == norm_name('cat', '**:*')


def test_norm_name_rel_ref_sub():
    """Test norm_name rel ref to a sibling module."""
    assert 'dog:bark' == norm_name('cat', '../dog:bark')
    assert 'dog:bark' == norm_name('cat', '../dog/.:bark')
    assert 'dog:bark' == norm_name('cat', './../dog:bark')
    assert 'dog:bark' == norm_name('cat', '.././dog:bark')
    assert 'dog:*' == norm_name('cat', '../dog:*')
    assert 'dog/**:*' == norm_name('cat', '../dog/**:*')


def test_norm_name_escape_sandbox_error():
    """Test norm_name various sandbox-escaping scenarios."""
    for bad_name in [
            '///cat/foo:bar',
            '//../cat/foo:bar',
            '/etc/passwd:meow',
            '../..:paw']:
        with pytest.raises(ValueError) as excinfo:
            norm_name('cat', bad_name)
    assert 'must remain inside project sandbox' in str(excinfo.value)


def test_norm_name_unqualified_error():
    """Test norm_name when given unqualified target name."""
    with pytest.raises(ValueError) as excinfo:
        norm_name('cat', 'poops')
    assert ('Must provide fully-qualified target name (with `:\') to avoid '
            'possible ambiguity' in str(excinfo.value))


_HELLO_PROG_HASH = '41eb61d311e6b841d71f2230e7e62a94'
_PROTO_BUILDER = '2673afde9594dc5212450377dfc16b41'
_BOTH_HASHES = list(sorted([_HELLO_PROG_HASH, _PROTO_BUILDER]))


_EXP_JSON = """{
    "buildenv": [],
    "builder_name": "CppApp",
    "deps": [
        "%s",
        "%s"
    ],
    "flavor": null,
    "name": "app:hello-prog-app",
    "props": {
        "attempts": 1,
        "base_image": [
            "%s"
        ],
        "build_params": {},
        "build_user": null,
        "distro": {},
        "docker_labels": {},
        "env": {},
        "executable": {},
        "full_path_cmd": false,
        "image_name": null,
        "image_tag": "latest",
        "license": [],
        "main": [
            "%s"
        ],
        "packaging_params": {},
        "policies": [],
        "run_user": null,
        "runtime_params": {},
        "work_dir": "/usr/src/app"
    },
    "tags": []
}""" % (_BOTH_HASHES[0], _BOTH_HASHES[1], _PROTO_BUILDER, _HELLO_PROG_HASH)


@pytest.mark.usefixtures('in_proto_project')
def test_target_hash_and_json(basic_conf):
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['app:hello-prog-app']
    populate_targets_graph(build_context, basic_conf)
    assert (_HELLO_PROG_HASH ==
            build_context.targets['app:hello-prog'].hash(build_context))
    assert (_PROTO_BUILDER ==
            build_context.targets[':proto-builder'].hash(build_context))
    prog_app = build_context.targets['app:hello-prog-app']
    prog_app.compute_json(build_context)
    assert _EXP_JSON == prog_app.json(build_context)


@pytest.mark.usefixtures('in_proto_project')
def test_hashify_targets(basic_conf):
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['app:hello-prog-app']
    populate_targets_graph(build_context, basic_conf)
    assert (
        _BOTH_HASHES ==
        hashify_targets(['app:hello-prog', ':proto-builder'], build_context))


def test_hashify_files():
    files = [join(join('tests', 'data'), fname)
             for fname in ('empty', 'hello.txt', 'world.txt')]
    hashed_files = hashify_files(files)
    assert {
        'tests/data/empty': 'd41d8cd98f00b204e9800998ecf8427e',
        'tests/data/hello.txt': '910c8bc73110b0cd1bc5d2bcae782511',
        'tests/data/world.txt': '910c8bc73110b0cd1bc5d2bcae782511',
    } == hashed_files



# -*- coding: utf-8 -*-

# Copyright 2019 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
utilities for tests
~~~~~~~~~~~~~~~~~~~~~~~

:author: Dana Shamir
"""
import random

import networkx


def generate_random_dag(nodes, min_rank=0, max_rank=10, edge_prob=0.3):
    """Return a random DAG with nodes from `nodes`.

    and edges can only go from nodes[i] -> nodes[j] if i < j
    (guaranteeing DAGness).
    """
    g = networkx.DiGraph()
    g.add_nodes_from(nodes)
    for j in range(1, len(nodes)):
        rank = random.randint(min_rank, min(j, max_rank))
        g.add_edges_from((nodes[i], nodes[j])
                         for i in random.sample(range(j), k=rank)
                         if random.random() > edge_prob)
    return g



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-argument

"""
yabt Custom Installer Builder
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from collections import namedtuple
import os
from os.path import basename, isdir, join, relpath, splitext
import shutil
import tarfile
from urllib.parse import urlparse
from zipfile import ZipFile

import git
from git import InvalidGitRepositoryError, NoSuchPathError
import requests

from ..artifact import ArtifactType as AT
from ..extend import (
    PropType as PT, register_build_func, register_builder_sig,
    register_manipulate_target_hook)
from ..logging import make_logger
from ..target_utils import split_name
from ..utils import rmtree, yprint


logger = make_logger(__name__)


register_builder_sig(
    'CustomInstaller',
    [('script', PT.File),
     ('fetch', PT.list, None),
     ('local_data', PT.FileList, None),
     ('uri', PT.str, None),  # deprecated
     ('uri_type', PT.str, None),  # deprecated
     ])


CustomInstaller = namedtuple('CustomInstaller',
                             ['name', 'package', 'install_script'])
FetchDesc = namedtuple('FetchDesc', ['uri', 'type', 'name'])


KNOWN_ARCHIVES = frozenset(('.gz', '.bz2', '.tgz', '.zip'))


def guess_uri_type(uri: str, hint: str=None):
    """Return a guess for the URI type based on the URI string `uri`.

    If `hint` is given, it is assumed to be the correct type.
    Otherwise, the URI is inspected using urlparse, and we try to guess
    whether it's a remote Git repository, a remote downloadable archive,
    or a local-only data.
    """
    # TODO(itamar): do this better
    if hint:
        return hint
    norm_uri = uri.lower()
    parsed_uri = urlparse(norm_uri)
    if parsed_uri.path.endswith('.git'):
        return 'git'
    if parsed_uri.scheme in ('http', 'https'):
        ext = splitext(parsed_uri.path)[-1]
        if ext in KNOWN_ARCHIVES:
            return 'archive'
        return 'single'
    return 'local'


def gitfilter(tarinfo):
    """Filter function for tar.add, to filter out git internal stuff."""
    if basename(tarinfo.name) in ['.git', '.gitignore']:
        return None
    return tarinfo


def git_handler(unused_build_context, target, fetch, package_dir, tar):
    """Handle remote Git repository URI.

    Clone the repository under the private builder workspace (unless already
    cloned), and add it to the package tar (filtering out git internals).

    TODO(itamar): Support branches / tags / specific commit hashes
    TODO(itamar): Support updating a cloned repository
    TODO(itamar): Handle submodules?
    TODO(itamar): Handle force pulls?
    """
    target_name = split_name(target.name)
    # clone the repository under a private builder workspace
    repo_dir = join(package_dir, fetch.name) if fetch.name else package_dir
    try:
        repo = git.Repo(repo_dir)
    except (InvalidGitRepositoryError, NoSuchPathError):
        repo = git.Repo.clone_from(fetch.uri, repo_dir)
    assert repo.working_tree_dir == repo_dir
    tar.add(package_dir, arcname=target_name, filter=gitfilter)


def fetch_url(url, dest, parent_to_remove_before_fetch):
    """Helper function to fetch a file from a URL."""
    logger.debug('Downloading file {} from {}', dest, url)
    try:
        shutil.rmtree(parent_to_remove_before_fetch)
    except FileNotFoundError:
        pass
    os.makedirs(parent_to_remove_before_fetch)
    # TODO(itamar): Better downloading (multi-process-multi-threaded?)
    # Consider offloading this to a "standalone app" invoked with Docker
    resp = requests.get(url, stream=True)
    with open(dest, 'wb') as fetch_file:
        for chunk in resp.iter_content(chunk_size=32 * 1024):
            fetch_file.write(chunk)


def archive_handler(unused_build_context, target, fetch, package_dir, tar):
    """Handle remote downloadable archive URI.

    Download the archive and cache it under the private builer workspace
    (unless already downloaded), extract it, and add the content to the
    package tar.

    TODO(itamar): Support re-downloading if remote changed compared to local.
    TODO(itamar): Support more archive formats (currently only tarballs).
    """
    package_dest = join(package_dir, basename(urlparse(fetch.uri).path))
    package_content_dir = join(package_dir, 'content')
    extract_dir = (join(package_content_dir, fetch.name)
                   if fetch.name else package_content_dir)
    fetch_url(fetch.uri, package_dest, package_dir)

    # TODO(itamar): Avoid repetition of splitting extension here and above
    # TODO(itamar): Don't use `extractall` on potentially untrsuted archives
    ext = splitext(package_dest)[-1].lower()
    if ext in ('.gz', '.bz2', '.tgz'):
        with tarfile.open(package_dest, 'r:*') as src_tar:
            src_tar.extractall(extract_dir)
    elif ext in ('.zip',):
        with ZipFile(package_dest, 'r') as zipf:
            zipf.extractall(extract_dir)
    else:
        raise ValueError('Unsupported extension {}'.format(ext))
    tar.add(package_content_dir, arcname=split_name(target.name))


def fetch_file_handler(unused_build_context, target, fetch, package_dir, tar):
    """Handle remote downloadable file URI.

    Download the file and cache it under the private builer workspace
    (unless already downloaded), and add it to the package tar.

    TODO(itamar): Support re-downloading if remote changed compared to local.
    """
    dl_dir = join(package_dir, fetch.name) if fetch.name else package_dir
    fetch_url(fetch.uri,
              join(dl_dir, basename(urlparse(fetch.uri).path)),
              dl_dir)
    tar.add(package_dir, arcname=split_name(target.name))


def local_handler(build_context, target, fetch, package_dir, tar):
    pass


def get_installer_desc(build_context, target) -> tuple:
    """Return a target_name, script_name, package_tarball tuple for `target`"""
    workspace_dir = build_context.get_workspace('CustomInstaller', target.name)
    target_name = split_name(target.name)
    script_name = basename(target.props.script)
    package_tarball = '{}.tar.gz'.format(join(workspace_dir, target_name))
    return target_name, script_name, package_tarball


@register_build_func('CustomInstaller')
def custom_installer_builder(build_context, target):
    yprint(build_context.conf, 'Fetch custom installer package', target)
    if target.props.fetch and (target.props.uri or target.props.uri_type):
        raise AttributeError(
            '{}: `uri` & `uri_type` are deprecated - use `fetch` exclusively'
            .format(target.name))

    workspace_dir = build_context.get_workspace('CustomInstaller', target.name)
    # cleanup workspace
    if isdir(workspace_dir):
        rmtree(workspace_dir)
    os.makedirs(workspace_dir)

    target_name, script_name, package_tarball = get_installer_desc(
        build_context, target)

    logger.debug('Making custom installer package {}', package_tarball)
    handlers = {
        'git': git_handler,
        'archive': archive_handler,
        'single': fetch_file_handler,
        'local': local_handler,
    }

    def to_fetch_desc(fetch_arg):
        if isinstance(fetch_arg, dict):
            return FetchDesc(
                uri=fetch_arg['uri'], type=fetch_arg.get('type', None),
                name=fetch_arg.get('name', None))
        assert isinstance(fetch_arg, str)
        return FetchDesc(uri=fetch_arg, type=None, name='')

    fetches = ([to_fetch_desc(fetch) for fetch in target.props.fetch]
               if target.props.fetch
               else [FetchDesc(name='', uri=target.props.uri,
                               type=target.props.uri_type)])
    if len(fetches) > 1 and any(fetch.name is None for fetch in fetches):
        raise ValueError('{}: Implicit empty name not allowed with multiple '
                         'fetches. To fetch a URI under the top dir, '
                         'explicitly set the name to \'\''.format(target.name))
    used_names = set()
    for fetch in fetches:
        if fetch.name in used_names:
            raise ValueError('{}: Duplicate fetch name "{}"'.format(
                target.name, fetch.name))
        used_names.add(fetch.name)
    tar = tarfile.open(package_tarball, 'w:gz', dereference=True)
    for fetch in fetches:
        uri_type = guess_uri_type(fetch.uri, fetch.type)
        logger.debug('CustomInstaller URI {} typed guessed to be {}',
                     fetch.uri, uri_type)
        package_dir = join(workspace_dir, target_name,
                           fetch.name or '_unnamed_')
        handlers[uri_type](build_context, target, fetch, package_dir, tar)
    # Add local data to installer package, if specified
    for local_node in target.props.local_data:
        tar.add(join(build_context.conf.project_root, local_node),
                arcname=join(target_name, local_node))
    # Add the install script to the installer package
    tar.add(join(build_context.conf.project_root, target.props.script),
            arcname=join(target_name, script_name))
    tar.close()
    target.artifacts.add(
        AT.custom_installer,
        relpath(package_tarball, build_context.conf.project_root))


@register_manipulate_target_hook('CustomInstaller')
def custom_installer_manipulate_target(build_context, target):
    target.tags.add('custom-installer')



import pytest
from subprocess import check_output

from yabt.buildcontext import BuildContext
from yabt.graph import populate_targets_graph

slow = pytest.mark.skipif(not pytest.config.getoption('--with-slow'),
                          reason='need --with-slow option to run')


@slow
@pytest.mark.usefixtures('in_cpp_project')
def test_file_group_dep(basic_conf):
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['hello:hello-with-data']
    populate_targets_graph(build_context, basic_conf)
    build_context.build_graph(run_tests=True)
    ls_output_image = str(check_output(
        ['docker', 'run', '--rm', '--entrypoint', 'ls',
         build_context.targets['hello:hello-with-data'].image_id,
         '-R', 'hello/data']))
    ls_output_here = str(check_output(['ls', '-R', 'hello/data']))
    assert ls_output_image == ls_output_here



# -*- coding: utf-8 -*-

# Copyright 2019 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt caching random tests
~~~~~~~~~~~~~~~~~~

:author: Dana Shamir
"""

import networkx as nx
import os
from os.path import join, dirname, abspath, getmtime
import pytest
import random
import shutil
import string

from conftest import reset_parser
from . import config, cli, extend
from .buildcontext import BuildContext
from .graph import populate_targets_graph
from .logging import make_logger
from .test_utils import generate_random_dag

NUM_TARGETS = 10
NUM_TESTS = 20

TMPL_DIR = join(dirname(abspath(__file__)), '..', 'tests', 'data',
                'caching')
CPP_TMPL = 'cpp_prog.cc.tmpl'
CPP_TEST_TMPL = 'cpp_test.cc.tmpl'
FAILING_TEST_TMPL = 'cpp_failing_test.cc.tmpl'
CPP_TARGET = """CppProg(
    '{}',
    sources='{}',
    in_buildenv=':builder', deps={}
)"""
CPP_TEST_TARGET = """CppGTest(
    '{}',
    sources='{}',
    in_buildenv=':builder-with-gtest',
    deps={}
)"""
YROOT_TMPL = 'YRoot.tmpl'
INSTALL_GTEST_SCRIPT = 'install-gtest.sh'
YSETTINGS = 'YSettings'

TARGET_TYPES = {
    'CppProg': (CPP_TMPL, CPP_TARGET),
    'CppGTest': (CPP_TEST_TMPL, CPP_TEST_TARGET)
}

GLOBAL_CACHE_DIR = '.global_cache'

slow = pytest.mark.skipif(not pytest.config.getoption('--with-slow'),
                          reason='need --with-slow option to run')

logger = make_logger(__name__)


class ProjectContext:
    def __init__(self):
        self.targets = {}
        self.test_targets = []
        self.targets_graph = None
        self.last_modified = {}
        self.last_run_tests = {}
        self.conf = None


def generate_random_project(size) -> ProjectContext:
    project = ProjectContext()
    targets = [random_string() for _ in range(size)]
    project.targets_graph = generate_random_dag(targets)
    for target in targets:
        target_type = random.choice(list(TARGET_TYPES.keys()))
        generate_file(target, target_type)
        project.targets[target] = target_type
        if target_type == 'CppGTest':
            project.test_targets.append(target)
    generate_yroot(project)
    with open(join(TMPL_DIR, YSETTINGS + '.tmpl'), 'r') as tmpl:
        ysettings = tmpl.read().format(GLOBAL_CACHE_DIR)
    with open(YSETTINGS, 'w') as ysettings_file:
        ysettings_file.write(ysettings)
    return project


def random_string():
    return ''.join([random.choice(string.ascii_letters + string.digits)
                    for _ in range(random.randint(20, 40))])


def generate_yroot(project: ProjectContext):
    yroot = []
    for target, target_type in project.targets.items():
        yroot.append(TARGET_TYPES[target_type][1].format(
            target, get_file_name(target),
            get_dependencies(target, project.targets_graph)))
    with open(join(TMPL_DIR, YROOT_TMPL), 'r') as yroot_tmpl_file:
        yroot_data = yroot_tmpl_file.read()
    with open(config.BUILD_PROJ_FILE, 'w') as yroot_file:
        yroot_file.write(yroot_data + '\n\n'.join(yroot))
    shutil.copyfile(join(TMPL_DIR, INSTALL_GTEST_SCRIPT), INSTALL_GTEST_SCRIPT)


def get_dependencies(target, target_graph):
    return [':' + dep for dep, other_target in target_graph.edges
            if other_target == target]


def generate_file(target, target_type, string_to_print=None):
    if string_to_print is None:
        string_to_print = target
    with open(join(TMPL_DIR, TARGET_TYPES[target_type][0]), 'r') as tmpl:
        code = tmpl.read().format(string_to_print)
    with open(get_file_name(target), 'w') as target_file:
        target_file.write(code)


def get_file_name(target):
    return join(target + '.cc')


def init_project(project):
    reset_parser()
    project.conf = cli.init_and_get_conf(['--non-interactive',
                                          '--continue-after-fail', 'build',
                                          '--upload-to-global-cache',
                                          '--download-from-global-cache'])
    extend.Plugin.load_plugins(project.conf)
    project.conf.targets = [':' + target for target in project.targets.keys()]
    build_context = BuildContext(project.conf)
    populate_targets_graph(build_context, project.conf)
    return build_context


def rebuild(project: ProjectContext):
    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    check_modified_targets(project, build_context, [])


def rebuild_after_modify(project: ProjectContext):
    target_to_change = random.choice(list(project.targets.keys()))
    logger.info('modifing target: {}'.format(target_to_change))
    generate_file(target_to_change, project.targets[target_to_change],
                  random_string())
    build_context = init_project(project)
    build_context.build_graph(run_tests=True)

    targets_to_build = nx.descendants(project.targets_graph, target_to_change)
    targets_to_build.add(target_to_change)
    check_modified_targets(project, build_context, targets_to_build)


def delete_file_and_return_no_modify(project: ProjectContext):
    target_to_delete = random.choice(list(project.targets.keys()))
    logger.info('deleting and returning the same for target: {}'
                .format(target_to_delete))
    file_name = get_file_name(target_to_delete)
    with open(file_name) as target_file:
        curr_content = target_file.read()
    os.remove(file_name)
    with open(file_name, 'w') as target_file:
        target_file.write(curr_content)

    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    check_modified_targets(project, build_context, [])


def add_dependency(project: ProjectContext):
    new_target = random_string()
    target_type = random.choice(list(TARGET_TYPES.keys()))
    logger.info('adding target: {} of type: {}'.format(new_target,
                                                       target_type))
    project.targets[new_target] = target_type
    project.last_modified[new_target] = None
    if target_type == 'CppGTest':
        project.test_targets.append(new_target)
        project.last_run_tests[new_target] = None
    project.conf.targets.append(':' + new_target)
    generate_file(new_target, target_type)

    project.targets_graph.add_node(new_target)
    targets = list(project.targets.keys())
    project.targets_graph.add_edges_from(
        (new_target, targets[i]) for i in range(len(targets) - 1)
        if random.random() > 0.8 and targets[i] != new_target)
    generate_yroot(project)

    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    targets_to_build = nx.descendants(project.targets_graph, new_target)
    targets_to_build.add(new_target)
    check_modified_targets(project, build_context, targets_to_build)


def download_from_global_cache(project: ProjectContext):
    target = random.choice(list(project.targets.keys()))
    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    check_modified_targets(project, build_context, [])
    cache_dir = project.conf.get_cache_dir(
        build_context.targets[':' + target], build_context)
    logger.info('removing cache from: {}'.format(cache_dir))
    shutil.rmtree(cache_dir)
    build_context.build_graph(run_tests=True)

    # We don't support globally caching tests yet
    check_modified_targets(project, build_context, [], [target])


def no_cache_at_all(project: ProjectContext):
    target_name = random.choice(list(project.targets.keys()))
    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    check_modified_targets(project, build_context, [])
    target = build_context.targets[':' + target_name]
    logger.info('removing local and global cache of target: {}'.format(
        target_name))
    shutil.rmtree(project.conf.get_cache_dir(target, build_context))
    shutil.rmtree(join(GLOBAL_CACHE_DIR, 'targets',
                       target.hash(build_context)))
    build_context.build_graph(run_tests=True)
    targets_to_build = nx.descendants(project.targets_graph, target_name)
    targets_to_build.add(target_name)
    check_modified_targets(project, build_context, targets_to_build)


def failing_test(project: ProjectContext):
    test_to_fail = random.choice(project.test_targets)
    logger.info('Making target: {} fail'.format(test_to_fail))
    with open(join(TMPL_DIR, FAILING_TEST_TMPL), 'r') as tmpl:
        code = tmpl.read().format(test_to_fail)
    with open(get_file_name(test_to_fail), 'w') as target_file:
        target_file.write(code)

    build_context = init_project(project)
    with pytest.raises(SystemExit):
        build_context.build_graph(run_tests=True)
    test_cache = get_test_cache(project.conf, test_to_fail, build_context)
    assert not os.path.isfile(test_cache), \
        "test: {} is failing but was put into cache".format(test_to_fail)

    generate_file(test_to_fail, project.targets[test_to_fail], random_string())
    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    targets_to_build = nx.descendants(project.targets_graph, test_to_fail)
    targets_to_build.add(test_to_fail)
    check_modified_targets(project, build_context, targets_to_build)


def check_modified_targets(project: ProjectContext, build_context,
                           targets_to_build, tests_to_build=None):
    for target, target_type in project.targets.items():
        last_modified = get_last_modified(project.conf, target, target_type)
        if target in targets_to_build:
            assert last_modified != project.last_modified[target], \
                "target: {} was supposed to be built again and" \
                " wasn't".format(target)
            project.last_modified[target] = last_modified
        else:
            assert last_modified == project.last_modified[target], \
                "target: {} was modified and it wasn't supposed" \
                " to".format(target)

    tests_to_build = tests_to_build or targets_to_build
    for target in project.test_targets:
        last_run = getmtime(get_test_cache(project.conf, target,
                                           build_context))
        if target in tests_to_build:
            assert last_run != project.last_run_tests[target],\
                "test: {} was supposed to run again but wasn't".format(target)
            project.last_run_tests[target] = last_run
        else:
            assert last_run == project.last_run_tests[target],\
                "test: {} was rerun and it wasn't supposed to".format(target)


def get_last_modified(basic_conf, target, target_type):
    return getmtime(join(basic_conf.builders_workspace_dir, 'release_flavor',
                         target_type, '_' + target, target + '.o'))


def get_test_cache(basic_conf, target, build_context):
    return join(basic_conf.get_cache_dir(build_context.targets[':' + target],
                                         build_context), 'tested.json')


@slow
def test_caching(tmp_dir):
    project = generate_random_project(NUM_TARGETS)

    build_context = init_project(project)
    build_context.build_graph(run_tests=True)
    logger.info('done first build')

    for target, target_type in project.targets.items():
        project.last_modified[target] = get_last_modified(project.conf, target,
                                                          target_type)
    for target in project.test_targets:
        project.last_run_tests[target] = getmtime(get_test_cache(
            project.conf, target, build_context))

    tests = [rebuild, rebuild_after_modify, delete_file_and_return_no_modify,
             add_dependency, failing_test, download_from_global_cache,
             no_cache_at_all]
    for i in range(NUM_TESTS):
        test_func = random.choice(tests)
        logger.info('starting build number: {} with func: {}'.format(
            i + 2, test_func.__name__))
        test_func(project)



# -*- coding: utf-8 -*-

# Copyright 2019 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A global cache implemented with google cloud storage
~~~~~~~~~~~~~~~~~~~

:author: Dana Shamir
"""

from google.cloud import storage
from os.path import join
from typing import List

from .global_cache import GlobalCache
from .logging import make_logger

SUMMARY_FILE = 'summary.json'
ARTIFACTS_FILE = 'artifact.json'
TARGETS_DIR = 'targets'
ARTIFACTS_DIR = 'artifacts'


logger = make_logger(__name__)


class GSGlobalCache(GlobalCache):
    def __init__(self, gce_project, bucket, directory=None):
        self.storage_client = storage.Client(gce_project)
        self.bucket = self.storage_client.get_bucket(bucket)
        self.targets_dir = join(directory, TARGETS_DIR) if directory else \
            TARGETS_DIR
        self.artifacts_dir = join(directory, ARTIFACTS_DIR) if directory \
            else ARTIFACTS_DIR

    def has_cache(self, target_hash: str):
        return self.bucket.blob(join(self.targets_dir, target_hash,
                                     SUMMARY_FILE)).exists()

    def download_summary(self, target_hash: str, dst: str):
        self.bucket.blob(join(self.targets_dir, target_hash,
                              SUMMARY_FILE)).download_to_filename(dst)

    def download_artifacts_meta(self, target_hash: str, dst: str):
        self.bucket.blob(join(self.targets_dir, target_hash,
                              ARTIFACTS_FILE)).download_to_filename(dst)

    def download_artifacts(self, artifacts_hashes: List[str], dst: str):
        if artifacts_hashes:
            # TODO(Dana): make this work in batch.
            # see https://github.com/googleapis/google-cloud-python/issues/3139
            for artifact_hash in artifacts_hashes:
                self.bucket.blob(join(self.artifacts_dir, artifact_hash))\
                    .download_to_filename(join(dst, artifact_hash))

    def create_target_cache(self, target_hash: str):
        pass

    def upload_summary(self, target_hash: str, src: str):
        self.bucket.blob(join(self.targets_dir, target_hash, SUMMARY_FILE))\
            .upload_from_filename(src)

    def upload_artifacts_meta(self, target_hash: str, src: str):
        self.bucket.blob(join(self.targets_dir, target_hash, ARTIFACTS_FILE))\
            .upload_from_filename(src)

    def upload_artifacts(self, artifacts_hashes: List[str], src: str):
        if artifacts_hashes:
            # TODO(Dana): make this work in batch
            for artifact_hash in artifacts_hashes:
                self.bucket.blob(join(self.artifacts_dir, artifact_hash))\
                    .upload_from_filename(join(src, artifact_hash))



# -*- coding: utf-8 -*-

# Copyright 2019 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A global cache implemented in local dick
~~~~~~~~~~~~~~~~~~~

:author: Dana Shamir
"""
import os
import shutil
from os.path import join, isdir
from typing import List

from .global_cache import GlobalCache

SUMMARY_FILE = 'summary.json'
ARTIFACTS_FILE = 'artifact.json'
TARGETS_DIR = 'targets'
ARTIFACTS_DIR = 'artifacts'


class FSGlobalCache(GlobalCache):
    def __init__(self, directory='/tmp/cache'):
        self.targets_dir = join(directory, TARGETS_DIR)
        self.artifacts_dir = join(directory, ARTIFACTS_DIR)
        os.makedirs(self.targets_dir, exist_ok=True)
        os.makedirs(self.artifacts_dir, exist_ok=True)

    def has_cache(self, target_hash: str):
        return isdir(join(self.targets_dir, target_hash))

    def download_summary(self, target_hash: str, dst: str):
        shutil.copyfile(join(self.targets_dir, target_hash, SUMMARY_FILE),
                        dst)

    def download_artifacts_meta(self, target_hash: str, dst: str):
        shutil.copyfile(join(self.targets_dir, target_hash, ARTIFACTS_FILE),
                        dst)

    def download_artifacts(self, artifacts_hashes: List[str], dst: str):
        for artifact_hash in artifacts_hashes:
            shutil.copyfile(join(self.artifacts_dir, artifact_hash),
                            join(dst, artifact_hash))

    def create_target_cache(self, target_hash: str):
        if not isdir(join(self.targets_dir, target_hash)):
            os.mkdir(join(self.targets_dir, target_hash))

    def upload_summary(self, target_hash: str, src: str):
        shutil.copyfile(src,
                        join(self.targets_dir, target_hash, SUMMARY_FILE))

    def upload_artifacts_meta(self, target_hash: str, src: str):
        shutil.copyfile(src,
                        join(self.targets_dir, target_hash, ARTIFACTS_FILE))

    def upload_artifacts(self, artifacts_hashes: List[str], src: str):
        for artifact_hash in artifacts_hashes:
            shutil.copyfile(join(src, artifact_hash),
                            join(self.artifacts_dir, artifact_hash))



# -*- coding: utf-8 -*-

# Copyright 2018 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt buildfile parser tests
~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""

from os.path import join

import pytest

from .buildcontext import BuildContext
from .graph import populate_targets_graph


@pytest.mark.usefixtures('in_error_project')
def test_parser_error(basic_conf, capsys):
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['parser-error']
    with pytest.raises(SystemExit):
        populate_targets_graph(build_context, basic_conf)
    _, err = capsys.readouterr()
    ybuild_path = join('tests', 'errors', 'parser-error', 'YBuild')
    assert '{}", line 6'.format(ybuild_path) in err
    assert (
        "Fatal: Must provide fully-qualified target name (with `:') to "
        "avoid possible ambiguity - `users' not valid\n" in err)



# -*- coding: utf-8 -*-

# Copyright 2018 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt policies tests
~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""

import pytest

from .buildcontext import BuildContext
from .graph import populate_targets_graph


@pytest.mark.usefixtures('in_error_project')
def test_policy_violation_unknown_license_name(basic_conf):
    basic_conf.targets = ['policy:bad-license-name']
    build_context = BuildContext(basic_conf)
    with pytest.raises(RuntimeError) as excinfo:
        populate_targets_graph(build_context, basic_conf)
    err_str = str(excinfo.value)
    assert ('Target policy:bad-license-name violates standard_licenses_only '
            'policy: Unknown license: GPLv3' in err_str)
    # asserting for 1 policy violation
    assert 2 == len(err_str.split('\n'))


@pytest.mark.usefixtures('in_error_project')
def test_policy_violation_bad_prod_license(basic_conf):
    basic_conf.targets = ['policy:app']
    build_context = BuildContext(basic_conf)
    with pytest.raises(RuntimeError) as excinfo:
        populate_targets_graph(build_context, basic_conf)
    err_str = str(excinfo.value)
    assert ('Target policy:app violates whitelist_prod_licenses policy: '
            'Invalid licenses for prod policy: GPL-3.0' in err_str)
    # asserting for 1 policy violation
    assert 2 == len(err_str.split('\n'))


@pytest.mark.usefixtures('in_error_project')
def test_no_violation(basic_conf):
    basic_conf.targets = ['policy:test']
    build_context = BuildContext(basic_conf)
    populate_targets_graph(build_context, basic_conf)
    # asserting no exception thrown


@pytest.mark.usefixtures('in_error_project')
def test_multiple_policy_violations(basic_conf):
    basic_conf.targets = ['policy']
    build_context = BuildContext(basic_conf)
    with pytest.raises(RuntimeError) as excinfo:
        populate_targets_graph(build_context, basic_conf)
    err_str = str(excinfo.value)
    assert ('Target policy:bad-license-name violates standard_licenses_only '
            'policy: Unknown license: GPLv3' in err_str)
    assert ('Target policy:app violates whitelist_prod_licenses policy: '
            'Invalid licenses for prod policy: GPL-3.0' in err_str)
    # asserting for 2 policy violations
    assert 3 == len(err_str.split('\n'))


@pytest.mark.usefixtures('in_error_project')
def test_disable_policy(nopolicy_conf):
    nopolicy_conf.targets = ['policy']
    build_context = BuildContext(nopolicy_conf)
    populate_targets_graph(build_context, nopolicy_conf)
    # asserting no exception thrown



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt target graph tests
~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from concurrent.futures import ThreadPoolExecutor
from functools import reduce
import random
from unittest.mock import Mock

import networkx
import pytest

from .test_utils import generate_random_dag
from .buildcontext import BuildContext
from .graph import (get_descendants, populate_targets_graph, topological_sort)


slow = pytest.mark.skipif(not pytest.config.getoption('--with-slow'),
                          reason='need --with-slow option to run')


def make_random_dag_build_context(
        num_nodes, min_rank=0, max_rank=10, edge_prob=0.3):
    """Return a build context based on a random DAG with `num_nodes` nodes."""

    class DummyTarget():
        def __init__(self, n):
            self.name = n
            self.value = n

    # Use random DAG to create a build context with dummy targets
    g = generate_random_dag(list(range(num_nodes)),
                            min_rank, max_rank, edge_prob)
    build_context = BuildContext(Mock())
    build_context.target_graph = g
    for n in g.nodes():
        build_context.targets[n] = DummyTarget(n)

    return g, build_context


def random_dag_scan(num_nodes):
    """Test that `target_iter` generates nodes in a correct order"""
    g, build_context = make_random_dag_build_context(num_nodes)
    # for every generated node, assert that all the prerequisite nodes are
    # already in the `done` set
    done = set()
    for target in build_context.target_iter():
        assert done.issuperset(get_descendants(g, target.name))
        target.done()
        done.add(target.name)
    # also assert that at the end, all nodes are "done"
    assert done == set(g.nodes())


def test_small_dag_scan():
    random_dag_scan(500)


@slow
def test_big_dag_scan():
    random_dag_scan(2000)


def multithreaded_dag_scanner(num_nodes, num_threads=16):
    """Test that a multi-threaded `target_iter` with `num_nodes` nodes and
       `num_threads` threads generates nodes in the correct order, by
       comparing reduce-friendly operations between the multi-threaded method
       and a topological-sort-based single-threaded method.
    """

    g, build_context = make_random_dag_build_context(num_nodes)
    # set leaf nodes values to random (-3,3) numbers,
    # and non-leaf nodes to "No value" (None)
    for n, out_deg in g.out_degree():
        build_context.targets[n].value = (random.randint(-3, 3)
                                          if out_deg == 0 else None)

    def func(target, values):
        """Reducer-friendly operator"""
        deps = list(g.successors(target.name))
        for dep in deps:
            assert values[dep] is not None
        if len(deps) == 0:
            # a leaf - just copy the value
            values[target.name] = target.value
        else:
            assert values[target.name] is None
            if len(deps) == 1:
                # node with one dependency - use dep value + 1
                values[target.name] = values[deps[0]] + 1
            else:
                # multi-dep node - apply reducer to dep values
                # reducer either sum or mult, depending on node parity
                reducer = ((lambda x, y: x + y) if target.name & 1
                           else (lambda x, y: x * y))
                values[target.name] = reduce(
                    reducer, (values[dep] for dep in deps))

    # Scan DAG in topological sort order, applying operator in order
    topo_vals = [None] * num_nodes
    for i in reversed(range(num_nodes)):
        func(build_context.targets[i], topo_vals)

    # Multi-threaded DAG scan using ready-queue
    queue_vals = [None] * num_nodes
    with ThreadPoolExecutor(max_workers=num_threads) as executor:

        def do_func(node):
            func(node, queue_vals)
            node.done()

        executor.map(do_func, build_context.target_iter())

    # Compare single vs. multi threaded results
    for topo_val, q_val in zip(topo_vals, queue_vals):
        assert topo_val is not None and q_val is not None
        assert topo_val == q_val


def test_small_multithreaded_dag_scan():
    multithreaded_dag_scanner(1000)


@slow
def test_big_multithreaded_dag_scan():
    multithreaded_dag_scanner(10000)


@pytest.mark.usefixtures('in_dag_project')
def test_target_graph(basic_conf):
    build_context = BuildContext(basic_conf)
    populate_targets_graph(build_context, basic_conf)
    assert (
        set(('yapi/server:users', ':flask', ':gunicorn', 'common:logging',
             'fe:fe', 'yapi/server:yapi', 'yapi/server:yapi-gunicorn',
             'common:base')) == set(build_context.target_graph.nodes))
    assert (
        set((('fe:fe', 'yapi/server:users'), ('fe:fe', ':flask'),
             ('fe:fe', 'common:base'), ('yapi/server:yapi', ':flask'),
             ('yapi/server:yapi', 'common:base'),
             ('yapi/server:yapi-gunicorn', 'yapi/server:yapi'),
             ('yapi/server:yapi-gunicorn', 'common:base'),
             ('yapi/server:yapi-gunicorn', ':gunicorn'),
             ('common:base', 'common:logging'))) ==
        set(build_context.target_graph.edges))
    assert ([':flask', ':gunicorn', 'common:logging', 'common:base',
             'yapi/server:users', 'fe:fe', 'yapi/server:yapi',
             'yapi/server:yapi-gunicorn'] ==
            list(topological_sort(build_context.target_graph)))


@pytest.mark.usefixtures('in_yapi_dir')
def test_target_graph_worldglob(basic_conf):
    """Test that building a graph with the world-glob specifier works."""
    basic_conf.targets = ['**:*']
    build_context = BuildContext(basic_conf)
    populate_targets_graph(build_context, basic_conf)
    assert (
        set(('yapi/server:users', ':flask', ':gunicorn', 'common:logging',
             'fe:fe', 'yapi/server:yapi', 'yapi/server:yapi-gunicorn',
             'common:base')) == set(build_context.target_graph.nodes))
    assert (
        set((('fe:fe', 'yapi/server:users'), ('fe:fe', ':flask'),
             ('fe:fe', 'common:base'), ('yapi/server:yapi', ':flask'),
             ('yapi/server:yapi', 'common:base'),
             ('yapi/server:yapi-gunicorn', 'yapi/server:yapi'),
             ('yapi/server:yapi-gunicorn', 'common:base'),
             ('yapi/server:yapi-gunicorn', ':gunicorn'),
             ('common:base', 'common:logging'))) ==
        set(build_context.target_graph.edges))


@pytest.mark.usefixtures('in_yapi_dir')
def test_target_graph_intenral_dir(basic_conf):
    """Test that building graph from internal dir works as expected."""
    basic_conf.targets = ['server']
    build_context = BuildContext(basic_conf)
    populate_targets_graph(build_context, basic_conf)
    assert (
        set(('yapi/server:users', ':flask', ':gunicorn', 'common:logging',
             'yapi/server:yapi', 'yapi/server:yapi-gunicorn',
             'common:base')) == set(build_context.target_graph.nodes))
    assert (
        set((('yapi/server:yapi', ':flask'),
             ('yapi/server:yapi', 'common:base'),
             ('yapi/server:yapi-gunicorn', 'yapi/server:yapi'),
             ('yapi/server:yapi-gunicorn', 'common:base'),
             ('yapi/server:yapi-gunicorn', ':gunicorn'),
             ('common:base', 'common:logging'))) ==
        set(build_context.target_graph.edges))


def test_stable_topological_sort():
    """Test that my modified topological sort is stable.

    Note - not doing many cycles because I saw that even with the non-stable
    implementation, it is stable in the context of the same process...
    (only reruns showed the unstable behavior)
    """
    expected_order = ['world', 'bar', 'baz', 'hello', 'foo']

    graph = networkx.DiGraph()
    graph.add_edges_from([('foo', 'bar'), ('foo', 'baz'),
                          ('bar', 'world'), ('foo', 'hello')])
    assert list(topological_sort(graph)) == expected_order

    same_graph = networkx.DiGraph({'foo': ['baz', 'hello', 'bar']})
    same_graph.add_edge('bar', 'world')
    assert list(topological_sort(same_graph)) == expected_order


def test_topological_sort1():
    graph = networkx.DiGraph()
    graph.add_edges_from([(1, 2), (1, 3), (2, 3)])
    assert list(topological_sort(graph)) == [3, 2, 1]

    graph.add_edge(3, 2)
    with pytest.raises(networkx.NetworkXUnfeasible):
        list(topological_sort(graph))

    graph.remove_edge(2, 3)
    assert list(topological_sort(graph)) == [2, 3, 1]


def test_topological_sort2():
    graph = networkx.DiGraph({1: [2], 2: [3], 3: [4],
                              4: [5], 5: [1], 11: [12],
                              12: [13], 13: [14], 14: [15]})

    with pytest.raises(networkx.NetworkXUnfeasible):
        list(topological_sort(graph))

    graph.remove_edge(1, 2)
    assert list(topological_sort(graph)) == [1, 5, 4, 3, 2, 15, 14, 13, 12, 11]


def test_topological_sort4():
    graph = networkx.Graph()
    graph.add_edge(1, 2)
    with pytest.raises(networkx.NetworkXError):
        list(topological_sort(graph))


def test_topological_sort4():
    graph = networkx.DiGraph()
    graph.add_edge(1, 2)
    assert list(topological_sort(graph)) == [2, 1]


@pytest.mark.usefixtures('in_error_project')
def test_graph_cycles(basic_conf):
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['cycle']
    with pytest.raises(RuntimeError) as excinfo:
        populate_targets_graph(build_context, basic_conf)
    ex_msg = str(excinfo.value)
    assert 'Detected cycles in build graph!' in ex_msg
    # expecting 3 cycles (so error message will have 4 lines)
    assert 4 == len(ex_msg.split('\n'))


@pytest.mark.usefixtures('in_error_project')
def test_dep_name_typo(basic_conf):
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['typo', 'typo:bar']
    with pytest.raises(ValueError) as excinfo:
        populate_targets_graph(build_context, basic_conf)
    ex_msg = str(excinfo.value)
    assert 'Could not resolve 5 targets' in ex_msg
    assert ':builderz - buildenv of typo:foo' in ex_msg
    assert 'typo:bar - seen on command line' in ex_msg
    assert 'typo:blask - dependency of typo:yapi' in ex_msg
    assert 'typo:loggin - dependency of typo:base' in ex_msg
    assert 'typo:zapi - dependency of typo:foo' in ex_msg
    # # expecting 5 unresolved targets (so error message will have 6 lines)
    assert 6 == len(ex_msg.split('\n'))



# -*- coding: utf-8 -*-

# Copyright 2018 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

:author: Itamar Ostricher
"""


from os.path import join
import shutil
from subprocess import check_output

import pytest

from . import cpp
from ..buildcontext import BuildContext
from ..graph import populate_targets_graph


slow = pytest.mark.skipif(not pytest.config.getoption('--with-slow'),
                          reason='need --with-slow option to run')


def clear_bin():
    try:
        shutil.rmtree('ybt_bin')
    except FileNotFoundError:
        pass


@pytest.mark.usefixtures('in_cpp_project')
@pytest.mark.parametrize(
    'test_case',
    (
        ('compiler_config:defaults', 'clang++-5.0',
         ['-std=c++14', '-Wall', '-fcolor-diagnostics', '-O2', '-DDEBUG'], []),
        ('compiler_config:override-compiler', 'foobar',
         ['-std=c++14', '-Wall', '-fcolor-diagnostics', '-O2', '-DDEBUG'], []),
        ('compiler_config:override-flags', 'clang++-5.0', ['-foo', 'bar'], []),
        # TODO: known failure - make it work...
        # ('compiler_config:override-flags-empty', 'clang++-5.0', [], []),
        ('compiler_config:post-extend-flags', 'clang++-5.0',
         ['-std=c++14', '-Wall', '-fcolor-diagnostics',
          '-O2', '-DDEBUG', '-foo', 'bar'], []),
        ('compiler_config:pre-extend-flags', 'clang++-5.0',
         ['-foo', 'bar', '-std=c++14', '-Wall', '-fcolor-diagnostics',
          '-O2', '-DDEBUG'], []),
        ('compiler_config:dep-extend-flags', 'clang++-5.0',
         ['-std=c++14', '-Wall', '-fcolor-diagnostics',
          '-O2', '-DDEBUG', '-DFOO=BAR'], ['-lfoo']),
    ))
def test_compiler_config(basic_conf, test_case):
    target_name, exp_compiler, exp_compile_flags, exp_link_flags = test_case
    build_context = BuildContext(basic_conf)
    basic_conf.targets = [target_name]
    populate_targets_graph(build_context, basic_conf)
    target = build_context.targets[target_name]
    cc = cpp.CompilerConfig(build_context, target)
    assert cc.compiler == exp_compiler
    assert cc.compile_flags == exp_compile_flags
    assert cc.link_flags == exp_link_flags
    # also check flavored workspace dir
    assert (build_context.get_workspace('foo', 'bar:baz') ==
            join(basic_conf.project_root,
                 'yabtwork', 'release_flavor', 'foo', 'bar_baz'))


@pytest.mark.usefixtures('in_cpp_project')
@pytest.mark.parametrize(
    'test_case',
    (
        ('compiler_config:defaults', 'clang++-5.0',
         ['-std=c++14', '-Wall', '-fcolor-diagnostics', '-g', '-DDEBUG']),
        ('compiler_config:override-compiler', 'foobar',
         ['-std=c++14', '-Wall', '-fcolor-diagnostics', '-g', '-DDEBUG']),
        ('compiler_config:override-flags', 'clang++-5.0', ['-foo', 'bar']),
        # TODO: known failure - make it work...
        # ('compiler_config:override-flags-empty', 'clang++-5.0', []),
        ('compiler_config:post-extend-flags', 'clang++-5.0',
         ['-std=c++14', '-Wall', '-fcolor-diagnostics',
          '-g', '-DDEBUG', '-foo', 'bar']),
        ('compiler_config:pre-extend-flags', 'clang++-5.0',
         ['-foo', 'bar', '-std=c++14', '-Wall', '-fcolor-diagnostics',
          '-g', '-DDEBUG']),
    ))
def test_compiler_config_debug(debug_conf, test_case):
    target_name, exp_compiler, exp_flags = test_case
    build_context = BuildContext(debug_conf)
    debug_conf.targets = [target_name]
    populate_targets_graph(build_context, debug_conf)
    target = build_context.targets[target_name]
    cc = cpp.CompilerConfig(build_context, target)
    assert cc.compiler == exp_compiler
    assert cc.compile_flags == exp_flags
    # also check flavored workspace dir
    assert (build_context.get_workspace('foo', 'bar:baz') ==
            join(debug_conf.project_root,
                 'yabtwork', 'debug_flavor', 'foo', 'bar_baz'))


@slow
@pytest.mark.parametrize(
    'target_name',
    ('hello:hello-app', 'hello_lib:hello-app', 'hello_mod/main:hello-app'))
@pytest.mark.usefixtures('in_cpp_project')
def test_cpp_builder(basic_conf, target_name):
    clear_bin()
    build_context = BuildContext(basic_conf)
    basic_conf.targets = [target_name]
    populate_targets_graph(build_context, basic_conf)
    build_context.build_graph()
    hello_out = str(
        check_output(['ybt_bin/{}'.format(target_name.replace(':', '/'))]))
    assert 'Hello world' in hello_out
    clear_bin()


@slow
@pytest.mark.usefixtures('in_tests_project')
def test_cpp_tester_success(basic_conf):
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['hello_gtest:greet-test']
    populate_targets_graph(build_context, basic_conf)
    build_context.build_graph(run_tests=True)


@slow
@pytest.mark.usefixtures('in_tests_project')
def test_cpp_tester_fail(basic_conf):
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['hello_gtest:greet-failing-test']
    populate_targets_graph(build_context, basic_conf)
    with pytest.raises(SystemExit):
        build_context.build_graph(run_tests=True)



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt ProtoBuf builder
~~~~~~~~~~~~~~~~~~~~~

:author: Zohar Rimon, Itamar Ostricher
"""


import os
from os.path import dirname, isfile, join, relpath, splitext
from pathlib import Path, PurePath

from ostrich.utils.path import commonpath

from ..artifact import ArtifactType as AT
from ..compat import walk
from ..extend import (
    PropType as PT, register_build_func, register_builder_sig,
    register_manipulate_target_hook)
from ..logging import make_logger
from ..utils import link_files, rmtree, yprint


logger = make_logger(__name__)


register_builder_sig(
    'Proto',
    [('sources', PT.FileList),
     ('in_buildenv', PT.Target),
     ('proto_cmd', PT.list, 'protoc'),
     ('gen_python', PT.bool, True),
     ('gen_cpp', PT.bool, True),
     ('gen_python_rpcz', PT.bool, False),
     ('gen_cpp_rpcz', PT.bool, False),
     ('gen_python_grpc', PT.bool, False),
     ('gen_cpp_grpc', PT.bool, False),
     ('grpc_plugin_path', PT.str, '/usr/local/bin/grpc_cpp_plugin'),
     ('copy_generated_to', PT.File, None),
     ('cmd_env', None),
     ])


@register_build_func('Proto')
def proto_builder(build_context, target):
    yprint(build_context.conf, 'Build ProtoBuf', target)
    workspace_dir = build_context.get_workspace('ProtoBuilder', target.name)
    # make sure workdir is clean
    rmtree(workspace_dir)
    proto_dir = join(workspace_dir, 'proto')
    # Collect proto sources from this target and all dependecies,
    # to link under the target sandbox dir
    protos = [source for source in target.props.sources
              if source.endswith('.proto')]
    for dep in build_context.generate_all_deps(target):
        if 'sources' in dep.props:
            protos.extend(source for source in dep.props.sources
                          if source.endswith('.proto'))
    link_files(protos, proto_dir, None, build_context.conf)
    buildenv_workspace = build_context.conf.host_to_buildenv_path(
        workspace_dir)
    protoc_cmd = target.props.proto_cmd + ['--proto_path', buildenv_workspace]
    if target.props.gen_cpp:
        protoc_cmd.extend(('--cpp_out', buildenv_workspace))
    if target.props.gen_python:
        protoc_cmd.extend(('--python_out', buildenv_workspace))
    if target.props.gen_cpp_grpc:
        protoc_cmd.extend(
            ('--grpc_out', buildenv_workspace,
             '--plugin=protoc-gen-grpc={}'
             .format(target.props.grpc_plugin_path)))
    if target.props.gen_python_grpc:
        protoc_cmd.extend(('--grpc_python_out', buildenv_workspace))
    if target.props.gen_cpp_rpcz:
        protoc_cmd.extend(('--cpp_rpcz_out', buildenv_workspace))
    if target.props.gen_python_rpcz:
        protoc_cmd.extend(('--python_rpcz_out', buildenv_workspace))
    protoc_cmd.extend((PurePath(buildenv_workspace) / 'proto' / src).as_posix()
                      for src in target.props.sources)
    build_context.run_in_buildenv(
        target.props.in_buildenv, protoc_cmd, target.props.cmd_env)
    generated_files = []

    def process_generated(gen_file: str, artifact_type: AT):
        if not isfile(gen_file):
            logger.error('Missing expected generated file: {}', gen_file)
            return
        target.artifacts.add(
            artifact_type,
            relpath(gen_file, build_context.conf.project_root),
            relpath(gen_file, workspace_dir))
        generated_files.append(gen_file)

    def create_init_py(path: str):
        init_py_path = join(path, '__init__.py')
        Path(init_py_path).touch(exist_ok=True)
        return init_py_path

    # Add generated files to artifacts / generated list
    for src in target.props.sources:
        src_base = join(proto_dir, splitext(src)[0])
        if target.props.gen_python:
            process_generated(src_base + '_pb2.py', AT.gen_py)
        if target.props.gen_cpp:
            process_generated(src_base + '.pb.cc', AT.gen_cc)
            process_generated(src_base + '.pb.h', AT.gen_h)
        if target.props.gen_python_rpcz:
            process_generated(src_base + '_rpcz.py', AT.gen_py)
        if target.props.gen_cpp_rpcz:
            process_generated(src_base + '.rpcz.cc', AT.gen_cc)
            process_generated(src_base + '.rpcz.h', AT.gen_h)
        if target.props.gen_python_grpc:
            process_generated(src_base + '_pb2_grpc.py', AT.gen_py)
        if target.props.gen_cpp_grpc:
            process_generated(src_base + '.grpc.pb.cc', AT.gen_cc)
            process_generated(src_base + '.grpc.pb.h', AT.gen_h)

    # Create __init__.py files in all generated directories with Python files
    if target.props.gen_python or target.props.gen_python_rpcz:
        py_dirs = set(('',))
        for src in target.props.sources:
            py_dir = dirname(src)
            while py_dir not in py_dirs:
                py_dirs.add(py_dir)
                py_dir = dirname(py_dir)
        for py_dir in py_dirs:
            process_generated(create_init_py(join(proto_dir, py_dir)),
                              AT.gen_py)

    # Copy generated files to external destination
    # TODO: remove this? (deprecate copy_generated_to?)
    if target.props.copy_generated_to:
        link_files(generated_files, target.props.copy_generated_to,
                   workspace_dir, build_context.conf)


def add_gen_python_path(target):
    if target.props.gen_python or target.props.gen_python_rpcz:
        join_env = target.props.packaging_params.pop('semicolon_join_env', {})
        if 'PYTHONPATH' in join_env:
            if '/usr/src/gen' not in join_env['PYTHONPATH'].split(':'):
                join_env['PYTHONPATH'] += ':/usr/src/gen'
        else:
            join_env['PYTHONPATH'] = '/usr/src/gen'
        target.props.packaging_params['semicolon_join_env'] = join_env


@register_manipulate_target_hook('Proto')
def proto_manipulate_target(build_context, target):
    target.buildenv = target.props.in_buildenv
    # manipulating the packaging_params (and storing it in the target props)
    # during target extraction (as opposed to build time), because:
    # 1. it should affect the target hash (for cache)
    # 2. it is used by docker builder further down the graph, so it should be
    #    populated if the target is cached (and build func is never called)
    add_gen_python_path(target)



# -*- coding: utf-8 -*-

# Copyright 2019 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
interface for a global cache
~~~~~~~~~~~~~~~~~~~

:author: Dana Shamir
"""

from typing import List


class GlobalCache:
    """
    An interface for a remote cache
    """
    def has_cache(self, target_hash: str) -> bool:
        """
        returns True if the cache has a target matching the hash
        """
        raise NotImplemented('Method has_cache of class {} was not '
                             'implemented'.format(self.__class__.__name__))

    def download_summary(self, target_hash: str, dst: str):
        """
        Download the summary file of the target to `dst`
        """
        raise NotImplemented('Method download_summary of class {} was not '
                             'implemented'.format(self.__class__.__name__))

    def download_artifacts_meta(self, target_hash: str, dst: str):
        """
        Downloads the metadata about the artifcats cached for the target to
        `dst`.
        """
        raise NotImplemented(
            'Method download_artifacts_meta of class {} was not '
            'implemented'.format(self.__class__.__name__))

    def download_artifacts(self, artifacts_hashes: List[str], dst: str):
        """
        Downloads the artifacts to `dst`. In `dst` each artifact will be in a
        file named with its hash.
        """
        raise NotImplemented('Method download_artifacts of class {} was not '
                             'implemented'.format(self.__class__.__name__))

    def create_target_cache(self, target_hash: str):
        raise NotImplemented('Method create_target_cache of class {} was not '
                             'implemented'.format(self.__class__.__name__))

    def upload_summary(self, target_hash: str, src: str):
        """
        Upload summary file in `src` describing the target with the given hash
        """
        raise NotImplemented('Method upload_summary of class {} was not '
                             'implemented'.format(self.__class__.__name__))

    def upload_artifacts_meta(self, target_hash: str, src: str):
        """
        Upload the metadata file in `src` describing the artifcats cached for
        the target.
        """
        raise NotImplemented(
            'Method upload_artifacts_meta of class {} was not '
            'implemented'.format(self.__class__.__name__))

    def upload_artifacts(self, artifacts_hashes: List[str], src: str):
        """
        Upload the artifacts with given hashes in directory `src`.
        """
        raise NotImplemented('Method upload_artifacts of class {} was not '
                             'implemented'.format(self.__class__.__name__))



# -*- coding: utf-8 -*-

# Copyright 2018 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt utils tests
~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from os.path import join

import pytest

from .utils import hash_tree, hash_file


DATA_DIR = join('tests', 'data')


def test_hashing_files():
    hello_hash_tree = hash_tree(join(DATA_DIR, 'hello.txt'))
    world_hash_tree = hash_tree(join(DATA_DIR, 'world.txt'))
    hello_hash_file = hash_file(join(DATA_DIR, 'hello.txt'))
    world_hash_file = hash_file(join(DATA_DIR, 'world.txt'))
    assert hello_hash_tree == world_hash_tree
    assert hello_hash_tree == hello_hash_file
    assert world_hash_tree == world_hash_file
    assert hello_hash_tree == '910c8bc73110b0cd1bc5d2bcae782511'


def test_hashing_empty_file():
    empty_hash = hash_tree(join(DATA_DIR, 'empty'))
    assert empty_hash == 'd41d8cd98f00b204e9800998ecf8427e'


def test_hashing_dirs():
    # test data includes non-ASCII characters in content as well as file names
    top1_hash = hash_tree(join(DATA_DIR, 'top1'))
    top2_hash = hash_tree(join(DATA_DIR, 'top2'))
    top3_hash = hash_tree(join(DATA_DIR, 'top3'))
    # this also verifies that the dir hash is not dependent upon the absolute
    # path of the dir itself
    assert top1_hash == top2_hash
    # same files but different sub-dir name should result different hash
    assert top1_hash != top3_hash
    assert top1_hash == '22f04e1ed792828b7053542cf7ce0448'
    assert top3_hash == '6f510185466f003a25bc249921e39b4d'
